const LRU = require('lru-cache');

function CacheLoader(options) {
  if (!options)
    options = {max: 1000, maxAge: 1000 * 60 * 60}; // default values
  let _cache = new LRU(options);
  let _hitCount = 0;
  let _missCount = 0;
  let _size = 0;

  let self = this;
  this.lazyForPromiseLoaders = function(key, time, loader, cb) {
    if (typeof time == 'function') {
      cb = loader;
      loader = time;
      time = undefined;
    }
    if (typeof loader !== 'undefined' && typeof loader !== 'function')
      throw new Error('Lazy loader value loader must be a function');

    if (typeof cb !== 'undefined' && typeof cb !== 'function') {
      throw new Error('Lazy loader callback must be a function');
    }
    return new Promise(async (resolve, reject) => {
      // checking if it is present in cache
      let value = this.get(key);
      if (value != null) {
        try {
          // the value itself could be a promise, handling for the same
          let promiseVal = await Promise.resolve(value);
          if (typeof cb === 'function') {
            cb({err: false, data: promiseVal});
          } else
            return resolve(promiseVal);
        } catch (e) {
          if (typeof cb === 'function') {
            cb({err: e, data: null});
          } else {
            reject(e);
          }
        }
        return;
      }

      // value is not present in cache, first will call the laoder function
      let promise = loader();
      // caching the promise in place of value, to serve same promises in case
      // before loader returns multiple calls are made
      this.put(key, promise, time);
      promise.then(result => {
        // now we got the actual result we will replace the previous promise with
        // actual result
        this.put(key, result, time);
        if (typeof cb === 'function')
          cb({err: false, data: result});
        else
          return resolve(result);
      }).catch(function(error) {
        self.del(key); // delete the promise from the key map as it errored out
        if (typeof cb === 'function') {
          cb({err: error, data: null});
        } else {
          reject(error);
        }
      });
    });
  };

  this.put = function(key, value, time) {
    if (typeof time !== 'undefined' &&
        (typeof time !== 'number' || isNaN(time) || time <= 0))
      throw new Error('Cache expiry timeout must be a positive number');

    if (!_cache.get(key))
      _size++;

    _cache.set(key, value, time);
    return value;
  };

  this.del = function(key) {
    let canDelete = true;

    if (!_cache.get(key))
      canDelete = false;

    if (canDelete)
      _del(key);

    return canDelete;
  };

  function _del(key) {
    _size--;
    _cache.del(key);
  }

  this.clear = function() {
    _size = 0;
    _cache.reset();
    _hitCount = 0;
    _missCount = 0;
  };

  this.get = function(key) {
    let data = _cache.get(key);
    if (data) {
      _hitCount++;
      return data;
    }

    _missCount++;
    return null;
  };

  this.size = function() {
    return _size;
  };

  this.hits = function() {
    return _hitCount;
  };

  this.misses = function() {
    return _missCount;
  };

  this.keys = function() {
    return _cache.keys();
  };

}

module.exports = CacheLoader;