function RedisCache(redisClient, options) {
  const prefix = options.PREFIX;
  const globalTTL = parseInt(options.TTL);
  if (!prefix) throw new Error('Prefix must be supplied');

  let _hitCount = 0;
  let _missCount = 0;

  // **helper functions
  const gKey = function(key) {
    return prefix + '{' + key + '}';
  };

  const lazifyLoaderHelper = async function(key, loader, options = {}) {
    const self = this;
    let value = await self.get(key);
    if (value) {
      _hitCount++;
      return value;
    }
    _missCount++;
    value = await loader();
    if (value == null) return null;
    await self.put(key, value, options.ttl);
    return value;
  };

  // **export functions
  this.lazyFetch = function(key, loader, options, cb) {
    if (typeof options === 'function') {
      cb = options;
      options = {};
    }
    if (typeof loader !== 'undefined' && typeof loader !== 'function')
      throw new Error('Lazy loader value loader must be a function');

    if (typeof cb === 'function') lazifyLoaderHelper.call(this, key, loader,
        options).then(r => cb(null, r)).
        catch(cb);
    else return lazifyLoaderHelper.call(this, key, loader, options);
  };

  this.put = function(key, value, ttlInSec, cb) {
    if (typeof ttlInSec === 'function') {
      cb = ttlInSec;
      ttlInSec = null;
    }
    let ttlIs = ttlInSec || globalTTL;
    if (typeof cb === 'function') {
      if (ttlIs) redisClient.setex(gKey(key), ttlIs, value, cb);
      else redisClient.set(gKey(key), value, cb);
    } else {
      if (ttlIs) return redisClient.setex(gKey(key), ttlIs, value);
      else return redisClient.set(gKey(key), value);
    }
  };

  this.del = (key, cb) => {
    if (typeof cb !== 'function') return redisClient.del(gKey(key));
    redisClient.del(gKey(key), cb);
  };

  this.get = (key, cb) =>{
    if (typeof cb !== 'function') return redisClient.get(gKey(key));
    redisClient.get(gKey(key), cb);
  };

  this.hits = () => _hitCount;

  this.misses = () => _missCount;
}

module.exports = RedisCache;
