const {Kafka} = require('kafkajs');
const Producer = require('./producer');
const Consumer = require('./consumer');
const busController = require('./busController');
const {v4: uuidv4} = require('uuid');
const utils = require('./utils');
const logger = utils.getLogger();
const CONTROL_EVENT_TOPIC = 'controlBusTopic';

function BusInterface(serviceName, config) {
  this._serviceName = serviceName;
  this._kafka = new Kafka(
      {clientId: serviceName, brokers: config.brokers, ssl: config.ssl});
  this.listOfConsumers = {};
  this.listOfProducers = {};
  this.listOfWrapConsumers = {};
  this.listOfWrapProducers = {};
}

BusInterface.prototype.initProducer = function(eventName, options, cb) {
  const self = this;
  if (typeof options === 'function') {
    cb = options;
    options = {};
  }
  self.listOfWrapProducers[eventName] = new Producer(this, eventName, options,
      (err, res) => {
        if (err) return cb(err, null);
        self.listOfProducers[eventName] = res.producer;
        cb(null, res.wrap);
      });
};

BusInterface.prototype.registerConsumer = function(
    sourceService, eventName, options, msgExec, cb) {
  const self = this;
  if (typeof options === 'function') {
    cb = msgExec;
    msgExec = options;
    options = {};
  }

  const topicName = utils.getTopicName(sourceService, eventName);
  self.listOfWrapConsumers[topicName] = new Consumer(this, topicName, options,
      msgExec, (err, res) => {
        if (err) return cb(err, null);
        self.listOfConsumers[topicName] = res.consumer;
        cb(null, res.wrap);
      });
};

BusInterface.prototype.startBusController = async function() {
  const self = this;
  const groupId = self._serviceName + '-controlConsumer' + '-' + uuidv4();
  const consumer = self._kafka.consumer({groupId});
  await consumer.connect();
  await consumer.subscribe({topic: CONTROL_EVENT_TOPIC, fromBeginning: false});
  await consumer.run({eachMessage: busController(self)});
  self._busConsumer = consumer;
};

BusInterface.prototype.close = function(cb) {
  const self = this;
  (async function() {
    await self._busConsumer.disconnect();
    for (let p of self.listOfProducers) await p.disconnect();
    for (let c of self.listOfConsumers) await c.disconnect();
    for (let k of Object.keys(self)) delete self[k];
  })().then(() => {
    logger.info('Successfully closed the Event Bus');
    if (typeof cb === 'function') cb(null);
  }).catch(e => {
    logger.error('Error in closing event bus', e);
    if (typeof cb === 'function') cb(e);
  });
};

module.exports = BusInterface;