const stringify = require('json-stringify');
const utils = require('./utils');
const logger = utils.getLogger();
const RateLimiter = require('limiter').RateLimiter;

/*
  ** NOTE**
  * Consumer class will be exposed out to client, so dont expose or leave
  * any direct access to kafka internal objects
 */
function Consumer(bus, topicName, options, msgExec, cb) {
  const kafka = bus._kafka;
  const self = this;
  // do not pass the client supplied config directly to the kafka module
  // use a wrapper and extract the required params
  const config = generateInitConfig(options);
  config.groupId = utils.getConsumerGroupName(topicName, bus._serviceName,
      options);
  // generate consumer
  const consumer = kafka.consumer(config);
  // establishing connection to kafka server
  consumer.connect().then(() => {
    // generate topic name
    consumer._topicName = topicName;
    const topic = {
      topic: consumer._topicName,
      fromBeginning: options.fromBeginning || false,
    };
    // subscribing to consumer
    consumer.subscribe(topic).then(() => {
      // get run config
      const config = generateSubscribeConfig(options);
      config.eachMessage = getAsyncEachMessageExecutor(options, msgExec);
      // attach
      consumer.run(config).then(() => {
        cb(null, {consumer, wrap: self});
      });
    });
  }).catch(error => {
    logger.error('Error initializing consumer for', topicName, error);
    cb(error, null);
  });

  this.pauseConsumption = () => {
    consumer.pause([{topic: consumer._topicName}]);
  };
  this.resumeConsumption = () => {
    consumer.resume([{topic: consumer._topicName}]);
  };
}

const getAsyncEachMessageExecutor = function(options, msgExec) {
  // default rate limit is 50 msgs per second
  let rateLimit = options.rateLimit || {rate: 20, interval: 'second'};
  const limiter = new RateLimiter(rateLimit.rate, rateLimit.interval);
  return function({topic, partition, message}) {
    return new Promise(resolve => {
      limiter.removeTokens(1, () => {
        messageExecutor(topic, partition, message, options, msgExec);
        resolve();
      });
    });
  };
};

const messageExecutor = function(topic, partition, message, options, msgExec) {
  if (logger.debug())
    logger.debug('Received message from topic', topic, 'partition', partition,
        'message', stringify(message));
  let headers = message.headers;
  for (let k of Object.keys(headers)) headers[k] = headers[k].toString();
  headers['timestamp'] = parseInt(headers['timestamp']);
  let key;
  if (message.key) key = message.key.toString();
  let value = message.value.toString();
  try {
    if (options.isJSONMsg) value = JSON.parse(value);
    if (key) msgExec(headers, value, key);
    else msgExec(headers, value);
  } catch (e) {
    logger.warn('Error executing msg exec', e);
    // continue here, but ideal scenario should push into an error queue
    // with retry mechanism in place
  }
};

const generateInitConfig = function(options) {
  return {};
};

const generateSubscribeConfig = function(options) {
  return {};
};

module.exports = Consumer;