const utils = require('./utils');
const logger = utils.getLogger();

const {CompressionTypes, CompressionCodecs} = require('kafkajs');
const SnappyCodec = require('kafkajs-snappy');
CompressionCodecs[CompressionTypes.Snappy] = SnappyCodec;

const HOSTNAME = require('os').hostname();

/*
  ** NOTE**
  * Producer class will be exposed out to client, so dont expose or leave
  * any direct access to kafka internal objects
 */
function Producer(bus, event, options, cb) {
  const kafka = bus._kafka;
  const self = this;
  //generate topic name
  let topicName;
  if (options.eventNameAsTopic)
    topicName = event;
  else topicName = utils.getTopicName(bus._serviceName, event);
  // do not pass the client supplied config directly to the kafka module
  // use a wrapper and extract the required params
  const initProducerConfig = generateInitConfig(options);
  // generate producer
  const producer = kafka.producer(initProducerConfig);
  // establish connection to kafka server
  producer.connect().then(() => {
    producer._topicName = topicName;
    //for producer nothing more to do
    cb(null, {producer, wrap: self});
  }).catch(error => {
    logger.error('Error initializing consumer for', topicName, error);
    cb(error, null);
  });
  this.send = generateSendMessageFunc(bus, producer, options, false);
  this.sendMulti = generateSendMessageFunc(bus, producer, options,
      true);
}

const generateSendMessageFunc = function(bus, producer, options, isMulti) {
  return function(key, msg, cb) {
    if (typeof cb !== 'function') cb = defaultSendMsgCb;
    // sending message
    producer.send({
      topic: producer._topicName,
      messages: formatMessage(key, msg, isMulti),
      acks: options.acks || 1,
      compression: CompressionTypes.Snappy,
    }).then(() => cb(null, msg)).catch(error => cb(error, null));
  };
};

const formatMessage = function(key, msg, isMulti) {
  const headers = {host: HOSTNAME, timestamp: Date.now().toString()};
  if (!isMulti) {
    let value;
    if (typeof msg === 'object')
      value = JSON.stringify(msg, JSONStringifyReplacer);
    else value = msg.toString();
    return [{key, value, headers}];
  } else {
    let messages = [];
    for (let m of msg) {
      let value;
      if (typeof m === 'object')
        value = JSON.stringify(m, JSONStringifyReplacer);
      else value = m.toString();
      messages.push({key, value, headers});
    }
    return messages;
  }
};

const defaultSendMsgCb = function(error, res) {
  if (error) return logger.error('Error sending message', error);
  logger.debug('Successfully sent the message', res);
};
const generateInitConfig = function(options) {
  return {
    retry: options.retry || {
      retries: 5,
    }
  };
};

const JSONStringifyReplacer = (key, value) => {
  if (value !== null) return value;
};

module.exports = Producer;