const loggerPack = require('./logger/logger');
const statsd = require('./statsd/statsd');
const requestPack = require('./request/request');
const restPack = require('./request/rest');
const redisPack = require('./redis/redis');

const CacheLoader = require('./cache/cacheLoader');
const RedisCacheLoader = require('./cache/redisCacheLoader');
const MongoGameLoader = require('./nosql/mongoGameLoader');
const CasualMongoGameLoader = require('./nosql/mongo-casual-game-collection')
const rds = require('./sql/rds');
const cassandra = require('./nosql/cassandra');
const mongo = require('./nosql/mongo');
const casualMongoReplicaSet = require('./nosql/mongo-casual-replica')
const casualHHMongo = require('./nosql/mongo-casual-hh')
const EventBus = require('./events');
const oddUtils = require('./utils/oddUtils');

const DEFAULT_CONFIG = require('./default_config');

let logger, SDC, requestHelper, restHelper, sessionRedis, gameRedis, localRedis,
    customRedis, gamesCollection, casualGamesCollection;

    const copyDefaults = function(config) {
  let defaultConfigs = oddUtils.deepClone(DEFAULT_CONFIG);
  if (typeof config.SERVICE_NAME === 'undefined')
    throw 'Service name should be compulsory. Check sample config file in the package.';
  defaultConfigs.SERVICE_NAME = config.SERVICE_NAME;
  if (typeof config.LUA_SCRIPTS_DIR !== 'undefined')
    defaultConfigs.LUA_SCRIPTS_DIR = config.LUA_SCRIPTS_DIR;
  Object.assign(defaultConfigs.LOGGER, config.LOGGER);
  Object.assign(defaultConfigs.STATSD, config.STATSD);
  if (config.REDIS && config.REDIS.SESSION_REDIS)
    defaultConfigs.REDIS.SESSION_REDIS = config.REDIS.SESSION_REDIS;
  if (config.REDIS && config.REDIS.GAME_REDIS)
    defaultConfigs.REDIS.GAME_REDIS = config.REDIS.GAME_REDIS;
  if (config.REDIS && config.REDIS.CUSTOM_REDIS)
    defaultConfigs.REDIS.CUSTOM_REDIS = config.REDIS.CUSTOM_REDIS;

  Object.assign(defaultConfigs.MONGO, config.MONGO);

  // assign required env variables
  process.env._SERVICE_NAME = config.SERVICE_NAME;
  process.env._LOG_LEVEL = defaultConfigs.LOGGER.LOG_LEVEL || 'info';
  return defaultConfigs;
};

exports.init = (config) => {
  let dConfig = copyDefaults(config);
  console.log('Using config ', JSON.stringify(dConfig, null, 2));
  logger = loggerPack(dConfig);
  return {
    getLogger: () => {
      return logger;
    },
    getStatsdClient: () => {
      if (typeof SDC === 'undefined')
        SDC = statsd(dConfig);
      return SDC;
    },
    getRequestHelper: () => {
      if (typeof requestHelper === 'undefined')
        requestHelper = requestPack(dConfig);
      return requestHelper;
    },
    getRestHelper: () => {
      if (typeof restHelper === 'undefined')
        restHelper = restPack(dConfig);
      return restHelper;
    },
    getSessionRedis: () => {
      if (typeof sessionRedis === 'undefined')
        sessionRedis = redisPack(dConfig, 'SESSION');
      return sessionRedis;
    },
    getGameRedis: () => {
      if (typeof gameRedis === 'undefined')
        gameRedis = redisPack(dConfig, 'GAME');
      return gameRedis;
    },
    getLocalRedis: () => {
      if (typeof localRedis === 'undefined')
        localRedis = redisPack(dConfig, 'LOCAL');
      return localRedis;
    },
    getCustomRedisCluster: () => {
      if (typeof customRedis === 'undefined')
        customRedis = redisPack(dConfig, 'CUSTOM');
      return customRedis;
    },
    getGamesCollection: () => {
      if (typeof gamesCollection === 'undefined')
        gamesCollection = new MongoGameLoader(dConfig);
      return gamesCollection;
    },
    getCasualGamesCollection: () => {
      if (typeof casualGamesCollection === 'undefined')
        casualGamesCollection = new CasualMongoGameLoader(config);
      return casualGamesCollection;
    }
  };
};

exports.utils = {
  CacheLoader,
  rds,
  EventBus,
  cassandra,
  RedisCacheLoader,
  oddUtils,
  mongo,
  casualMongoReplicaSet,
  casualHHMongo
};
