const redis = require("ioredis");
const fs = require("fs");
const path = require("path");
const async = require("async");
const loggerPack = require("../logger/logger");
const statsdPack = require("../statsd/statsd");

let logger, dir, conf, SDC;

/**
 * Function does 2 tasks.
 * 1. Loads lua scripts as redis commands to ioredis object.
 * 2. Wraps over the custom lua script objects to do custom commands like sending stats, etc.
 * @param cluster ioredis cluster object
 * @param wrapper Dummy object. ideally should be empty object.
 * @param callback
 */
const luaScriptLoader = function(cluster, wrapper, callback) {
   // Lua script loading and dummy orchestrating.
   fs.readdir(dir, function(error, files) {
      // getting the list of all files from the supplied lua script directory name in config.
      if (error) {
         logger.error("Could not list the directory for reading lua scripts.", error);
         callback(error, null);
         return;
      }
      // for each filename
      async.each(files, function(file, fn) {
         let fileName = file.slice(0, -4);
         logger.debug("Loading script", fileName);
         // loading the lua script as a command into ioredis object
         cluster.defineCommand(fileName, {
            numberOfKeys: 1,
            lua: fs.readFileSync(path.resolve(dir, file), {
               encoding: "utf8",
            }),
         });
         // dummy wrapping the lua functions with custom commands before and after execution of script
         wrapper[fileName] = function(...args) {
            let cb = args[args.length - 1];
            let params = args;
            if (typeof cb === "function") {
               params = args.slice(0, args.length - 1); // if it is a callback remove the last element
            }
            return new Promise(async function(resolve, reject) {
               try {
                  let startTime = +new Date();
                  let data = await cluster[fileName](...params);
                  let timeTaken = +new Date() - startTime;
                  if (conf.STATSD.SEND_LUA_SCRIPT_TIME)
                     SDC.timing("lua." + fileName, timeTaken);
                  resolve(data);
                  if (typeof cb === "function")
                     cb(null, data);
               } catch (err) {
                  if (conf.STATSD.SEND_LUA_SCRIPT_ERROR_STATS)
                     SDC.increment("lua_error." + fileName, 1);
                  if (typeof cb === "function") {
                     cb(err, null);
                     resolve(); // this if else should be removed and should always reject, if unhandled rejection is allowed.
                  } else
                     reject(err);
               }
            });
         };
         fn();
      }, function(error) {
         if (error) {
            logger.error("Failed to load lua scripts to redis cluster.");
            callback(error, null);
         } else {
            logger.info("Successfully loaded all lua scripts to redis cluster.");
            callback(null, "Success");
         }
      });
   });
};

/**
 * Reroutes default functions of cluster into dummy wrapper object i.e, dummy object becomes wrapper for cluster object.
 * Ex. set, get, hset, zadd, pipeline etc..
 * @param cluster ioredis cluster object
 * @param wrapper Dummy object. ideally should be empty object.
 */
const dummify = function(cluster, wrapper) {
   // Getting list of all function names in ioredis object
   let funcNames = [];
   for (let func in cluster) {
      if (typeof cluster[func] == "function")
         funcNames.push(func);
   }
   // for each function name in ioredis object creating dummy function in wrapper
   // that routes all calls to ioredis object.
   for (let func of funcNames) {
      wrapper[func] = function(...args) {
         return cluster[func](...args);
      };
   }
   // to support current code base practices
   wrapper.getPipeline = () => {
      return cluster.pipeline();
   };
};

module.exports = (config, clusterName) => {
   conf = config;
   let client;
   if (clusterName === 'GAME')
      client = new redis.Cluster(config.REDIS.GAME_REDIS);
   else if (clusterName === 'SESSION')
      client = new redis.Cluster(config.REDIS.SESSION_REDIS);
   else if (clusterName === 'LOCAL')
      client = new redis();
   else if (clusterName === 'CUSTOM')
      client = new redis.Cluster(config.REDIS.CUSTOM_REDIS);
   else throw 'Either GAME or SESSION is only supported now.';
   logger = loggerPack(config);
   SDC = statsdPack(config);
   dir = config.LUA_SCRIPTS_DIR;
   let wrapper = {};
   dummify(client, wrapper);
   if (dir) luaScriptLoader(client, wrapper, () => {
      logger.info('Successfully loaded all scripts.');
   });
   return wrapper;
};