const sdcPack = require("../statsd/statsd");
const restPack = require("./rest");
const bunyan = require("bunyan");

let SDC, restHelper;

let conf;

module.exports = (config) => {
   conf = config;
   SDC = sdcPack(config);
   return {
      requestHook, responseHook, loggerHook,
   };
};

const initGkLogger = function(req, service, level) {
   req.logger = {
      __requestId: null,
      __sessionId: null,
      __deviceId: null,
      __userId: null,

      __logger: null,

      __init: function(requestId) {
         this.__logger = bunyan.createLogger({
            name: service,
            serializers: bunyan.stdSerializers,
            stream: process.out,
            level: level,
            src: false,
         });

         this.__requestId = requestId;

         this.__logger.info = this.__logger.info.bind(this.__logger);
         this.__logger.error = this.__logger.error.bind(this.__logger);
         this.__logger.fatal = this.__logger.fatal.bind(this.__logger);
         this.__logger.debug = this.__logger.debug.bind(this.__logger);

      },

      setSessionId: function(sessionId) {
         this.__sessionId = sessionId;
      },
      setDeviceId: function(deviceId) {
         this.__deviceId = deviceId;
      },
      setUserId: function(userId) {
         this.__userId = userId;
      },
      setRequestId: function(requestId) {
         this.requestId = requestId;
      },
      logRequest: function(request) {
         if (request.body)
            this.info(
                '{ path:' + request.path + ', queryParams:' +
                JSON.stringify(request.query) + ', body: ' +
                JSON.stringify(request.body) + '}');
         else
            this.info('{ path:' + request.path + ', queryParams:' +
                JSON.stringify(request.query) +
                ', body: "Not logging the body"}');
      },
      logResponse: function(code, body) {
         if (body)
            this.info('{ responseCode: ' + code + ', resBody:' + body + '}');
         else
            this.info('{ responseCode: ' + code +
                ', resBody:"Not logging the body"}');

      },
      trace: function(data, ...message) {
         if (data == null && message.length === 0) return this.__logger.trace();
         this.__log('trace', data, message);
      },
      debug: function(data, ...message) {
         if (data == null && message.length === 0) return this.__logger.debug();
         this.__log('debug', data, message);
      },
      info: function(data, ...message) {
         if (data == null && message.length === 0) return this.__logger.info();
         this.__log('info', data, message);
      },
      warn: function(data, ...message) {
         if (data == null && message.length === 0) return this.__logger.warn();
         this.__log('warn', data, message);
      },
      error: function(data, ...message) {
         if (data == null && message.length === 0) return this.__logger.error();
         this.__log('error', data, message);
      },
      fatal: function(data, ...message) {
         if (data == null && message.length === 0) return this.__logger.fatal();
         this.__log('fatal', data, message);
      },

      __log(loggerMethod, data, message) {
         if (data == null) data = {};
         if ( message.length === 0)
            message = [];	// First argument was message

         if (typeof data !== 'object') {
            message.splice(0, 0, data);
            data = {};
         }

         data.requestId = this.__requestId;
         data.sessionId = this.__sessionId;
         data.deviceId = this.__deviceId;
         data.userId = this.__userId;

         this.__logger[loggerMethod](data, ...message);
      },
   };

   // Now we can initialize the logger
   req.logger.__init(req.requestId);
};

const requestHook = function(req, res, next) {
   if (conf.LOGGER.LOG_REQUEST)
      req.logger.logRequest(req);
   else {
      let body = req.body;
      req.body = undefined;
      req.logger.logRequest(req);
      req.body = body;
   }
   //statsd
   let pathStat = (req.method + req.path).toLowerCase().
      replace(/[:.]/g, "").
      replace(/\//g, "_");
   if (conf.STATSD.SEND_PATH_STATS)
      SDC.increment("path." + pathStat, 1);
   req.startTime = +new Date();
   next();
};

const loggerHook = function(request, response, next) {
   let headers = request.headers;
   request.requestId = headers["gk-request-id"];
   initGkLogger(request, conf.SERVICE_NAME, conf.LOGGER.LOG_LEVEL);
   request.logger.setRequestId(headers["gk-request-id"]);
   request.logger.setSessionId(headers["gk-session-id"]);
   request.logger.setDeviceId(headers["gk-device-id"]);
   request.logger.setUserId(headers["gk-user-id"]);
   // for backwards compatibility, hence this if-else block
   if (!request.internalRestClient) {
      if (typeof restHelper === "undefined") restHelper = restPack(conf);
      restHelper.initInternalRestClient(request, headers['gk-request-id'],
          headers['gk-session-id'], headers['gk-device-id'],
          headers['gk-user-id'], headers['x-forwarded-for']);
   }
   else
      request.internalRestClient.logger = request.logger;
   next();
};

const responseHook = function(req, res, next) {
   let oldWrite = res.write,
      oldEnd = res.end;
   let pathStat = (req.method + req.path).toLowerCase().
      replace(/[:.]/g, "").
      replace(/\//g, "_");
   let chunks = [];
   res.write = function(chunk) {
      chunks.push(Buffer.from(chunk));
      oldWrite.apply(res, arguments);
   };
   res.end = function(chunk) {
      if (chunk)
         chunks.push(Buffer.from(chunk));
      if (conf.LOGGER.LOG_RESPONSE) {
         let body = Buffer.concat(chunks).toString('utf8');
         req.logger.logResponse(res.statusCode, body);
      } else
         req.logger.logResponse(res.statusCode);

      // statsD
      let timeTaken = +new Date() - req.startTime;
      if (conf.STATSD.SEND_RESPONSE_CODE_STATS)
         SDC.increment('responseCode.' + pathStat + '.' + res.statusCode, 1);
      if (conf.STATSD.SEND_RESPONSE_TIME_STATS)
         SDC.timing("resTime." + pathStat, timeTaken);
      oldEnd.apply(res, arguments);
   };
   next();
};