const request = require("superagent");
const Agent = require("agentkeepalive");
const statsd = require('../statsd/statsd');
const {v4: uuidv4} = require('uuid');

const keepaliveAgent = new Agent({
});

let SDC;
let conf;

module.exports = (config) => {
   conf = config;
   SDC = statsd(config);
  require('./retryHelper')(request, config);
  return {initInternalRestClient};
};

const initInternalRestClient = function(
    req, requestId, sessionId, deviceId, userId, clientIp, timeout) {

   req.internalRestClient = {

     __clientIp: clientIp,
     __deviceId: deviceId,
     __sessionId: sessionId,
     __requestId: requestId,
     __userId: userId,
     logger: req.logger,
     __timeout: parseInt(timeout) || 3000,

     getRequestId: function() {
       return this.__requestId;
     },

     sendGetRequest: function(url, callback, timeout, retry, headers) {
       if (typeof callback !== 'function')
         return new Promise((resolve, reject) =>
             this.__sendRequest('GET', url, {}, timeout, retry, headers,(err, res) => {
               if (err) return reject(err);
               if (res.err) return reject(res.error || res.err);
               resolve(res);
             }));
       return this.__sendRequest('GET', url, {}, timeout, retry, headers,callback);
     },

     sendPostRequest: function(url, requestBody, callback, timeout, retry, headers) {
       if (typeof callback !== 'function')
         return new Promise((resolve, reject) =>
             this.__sendRequest('POST', url, requestBody, timeout, retry, headers,
                 (err, res) => {
                   if (err) return reject(err);
                   if (res.err) return reject(res.error || res.err);
                   resolve(res);
                 }));
       return this.__sendRequest(
           'POST', url, requestBody, timeout, retry, headers,callback);
     },

     sendPutRequest: function(url, requestBody, callback, timeout, retry, headers) {
       if (typeof callback !== 'function')
         return new Promise((resolve, reject) =>
             this.__sendRequest(
                 'PUT', url, requestBody, timeout, retry, headers, (err, res) => {
                   if (err) return reject(err);
                   if (res.err) return reject(res.error || res.err);
                   resolve(res);
                 }));
       return this.__sendRequest(
           'PUT', url, requestBody, timeout, retry, headers, callback);
     },

     __sendRequest: function(
         method, url, requestBody, timeout, retry = 3, headers, callback) {
       const self = this;
       const requestId = this.__requestId || uuidv4();

       if (conf.LOGGER.LOG_EXTERNAL_CALL_REQUEST)
         self.logger.info({method, url, requestBody, requestId});

       let req = request(method, url).
           agent(keepaliveAgent).
           set('Accept', 'application/json').
           set('gk-request-id', requestId).
           set('gk-session-id', !this.__sessionId ? '' : this.__sessionId).
           set('gk-device-id', !this.__deviceId ? '' : this.__deviceId).
           set('gk-user-id', !this.__userId ? '' : this.__userId).
           set('gk-client-ip', !this.__clientIp ? '' : this.__clientIp).
           set('gk-interaction-id', 1).
           retry(retry, 3000, [400, 401, 404]).
           timeout({
             deadline: timeout || this.__timeout,
             response: timeout || this.__timeout,
           });

       if (typeof headers === 'object')
         for (let [k, v] of Object.entries(headers)) req.set(k, v);
         // Now send the request
         if (requestBody.__FILES__) {

            // Multipart request
            // Note: send method and content type is not required in multipart requests.
            const files = requestBody.__FILES__;
            const fields = requestBody.__FIELDS__;

            if (conf.LOGGER.LOG_EXTERNAL_CALL_REQUEST)
               self.logger.info({fields: fields, files: files},
                  "This is a multipart request");

            Object.keys(files).forEach(function(param) {
               req.attach(param, files[param]);
            });

            // Remaining fields
            Object.keys(fields).forEach(function(name) {
               req.field(name, "" + fields[name]);	// Everything should be string
            });

         } else {
            req.set("Content-Type", "application/json").send(requestBody);
         }

         // statsd
         let pathStat = (method + url).split("?")[0].toLowerCase().
            replace(/[:.=?]/g, "").
            replace(/\//g, "_");
         req.startTime = +new Date();

         req.end(function(err, res) {
           // callback already called return
           if (req.__callbackAlCalled) return;
           req.__callbackAlCalled = true;

            let response = {};

            if (res) {
               response = res.body;
            }

            let error;
            if (err) {
              // if unable to get any response, most probably due to server not being available
              self.logger.error({
                    method: method,
                    url: url,
                    code: err.code,
                    message: err.message,
                  },
                  'Error in calling rest service. No response.');

              error = {
                  status: err.status,
                statusCode: res ? res.statusCode : '',
               };

               if (conf.STATSD.SEND_EXTERNAL_CALL_RESPONSE_ERRORS)
                 SDC.increment('extCall.' + pathStat + '.httpError');
            } else if (res && res.statusCode >= 400) {
              self.logger.error(
                  {method: method, url: url, statusCode: res.statusCode},
                  'Error in calling rest service. Error Http Status code observed');
              error = {statusCode: res.statusCode};

              if (conf.STATSD.SEND_EXTERNAL_CALL_RESPONSE_ERRORS)
                SDC.increment('extCall.' + pathStat + '.httpError');
            } else {
              if (response.err) { // got 200 OK but error in response, then log
                self.logger.error({
                  method: method,
                  url: url,
                  statusCode: res.statusCode,
                  response: response,
                }, 'Found error in response body.');
                if (conf.STATSD.SEND_EXTERNAL_CALL_RESPONSE_ERRORS)
                  SDC.increment('extCall.' + pathStat + '.resError');
              } else {
                if (conf.LOGGER.LOG_EXTERNAL_CALL_RESPONSE)
                  self.logger.info({
                    method: method,
                    url: url,
                    statusCode: res.statusCode,
                  });
                if (conf.STATSD.SEND_EXTERNAL_CALL_RESPONSE_SUCCESS)
                  SDC.increment('extCall.' + pathStat + '.resSuccess');
              }
            }
            let timeTaken = +new Date() - req.startTime;
            if (conf.STATSD.SEND_EXTERNAL_CALL_RESPONSE_TIME)
               SDC.timing("extCall." + pathStat + ".resTime", timeTaken);
            callback(error, response);
         });
      },
   };
};


