const mysql = require("mysql");

let mysqlConnectionPoolMap = {};

module.exports.initConnectionPool = function(host, db, user, password, poolSize, key) {
   if (!key)
      key = this.makeKey(host, db, user);

   if (mysqlConnectionPoolMap[key] == null) {
      mysqlConnectionPoolMap[key] = mysql.createPool({
         host: host,
         user: user,
         password: password,
         database: db,
         connectionLimit: poolSize,
      });
   }
};

module.exports.makeKey = function(host, db, user) {
   return host + "_" + db + "_" + user;
};

module.exports.getDbConnection = function(key) {
   let mysqlConnectionPool = mysqlConnectionPoolMap[key];

   if (!mysqlConnectionPool)
      throw new Error("Connection pool is not initialized.");

   const query = function(statement, values, fn, resultHandler) {
      let isCallback = false;
      if (typeof values === "function")
         fn = values;
      if (typeof fn === "function")
         isCallback = true;

      return new Promise((resolve, reject) => {
         mysqlConnectionPool.getConnection((err, mysqlConnection) => {
            if (!err) {
               function callback(err, record) {
                  if (!err) {
                     if (isCallback) fn({err: false, data: resultHandler(record)});
                     resolve(resultHandler(record));
                  } else {
                     if (isCallback) fn({err: true, errCode: "1007"});
                     else reject({err: err, errCode: "1007"});
                  }
                  mysqlConnection.release();
               }

               if (typeof values === "function")
                  mysqlConnection.query(statement, callback);
               else
                  mysqlConnection.query(statement, values, callback);

            } else {
               if (isCallback)
                  fn({err: true, errCode: "1006"});
               else
                  reject({err: err, errCode: "1006"});
            }
         });
      });
   };

   return {
      SELECT: function(statement, values, fn) {
         return query(statement, values, fn, (record) => {
            return record;
         });
      },
      UPDATE: function(statement, values, fn) {
         return query(statement, values, fn, (record) => {
            return record;
         });
      },
      INSERT: function(statement, values, fn) {
         return query(statement, values, fn, (record) => {
            return record.insertId;
         });
      },
      beginTransaction: function(fn) {
         let isCallback = false;
         if (typeof fn === "function")
            isCallback = true;

         return new Promise((resolve, reject) => {
            mysqlConnectionPool.getConnection((err, mysqlConnection) => {
               if (!err) {
                  mysqlConnection.beginTransaction((err) => {
                     if (!err) {
                        if (isCallback) fn({err: false, data: mysqlConnection});
                        resolve(mysqlConnection);
                     } else {
                        if (isCallback) fn({err: true, errCode: "2002"});
                        else reject({err: err, errCode: "2002"});
                     }
                  });
               } else {
                  if (isCallback) fn({err: true, errCode: "2001"});
                  else reject({err: err, errCode: "2001"});
               }
            });
         });
      },

      TransactionQuery: function(connection, statement, values, fn) {
         let isCallback = false;
         if (typeof values === "function")
            fn = values;
         if (typeof fn === "function")
            isCallback = true;

         return new Promise((resolve, reject) => {
            function callback(err, record) {
               if (!err) {
                  if (isCallback) fn({err: false, data: record});
                  resolve(record);
               } else {
                  if (isCallback) fn({err: true, errCode: "1007"});
                  else reject({err: err, errCode: "1007"});
               }
            }

            if (typeof values === "function")
               connection.query(statement, callback);
            else
               connection.query(statement, values, callback);
         });
      },

      COMMIT: function(connection, fn) {
         let isCallback = false;
         if (typeof fn === "function")
            isCallback = true;

         return new Promise((resolve, reject) => {
            connection.commit((err) => {
               connection.release();
               if (!err) {
                  if (isCallback) fn({err: false});
                  resolve();
               } else {
                  if (isCallback) fn({err: true, errCode: "2006"});
                  else reject({err: err, errCode: "2006"});
               }
            });
         });
      },

      ROLLBACK: function(connection) {
         connection.rollback();
         connection.release();
      },
   };
};